%%%
%%% Forward YDgCoCg-WSST by Taizo Suzuki
%%%
%%% Taizo Suzuki,
%%% "Wavelet-based spectral-spatial transforms for CFA-sampled raw camera image compression,"
%%% IEEE Trans. Image Process., Vol. 29, No. 1, pp. 433-444, Dec. 2020.
%%% DOI: 10.1109/TIP.2019.2928124
%%%
%%% [y, dg, co, cg] = ydgcocg_wsst(r, g1, g2, b, type)
%%%
%%% type
%%%  0: Malvar, DCC 2012 or Haar, YDgCoCg-WSST
%%% 53: 5/3, YDgCoCg-WSST
%%% 97: 9/7, YDgCoCg-WSST
%%%
function [y, dg, co, cg] = ydgcocg_wsst_forward(r, g1, g2, b, type)

%  r  | g1
% ----+----
%  g2 | b
switch type
    case 0 % Malvar, DCC 2012 or Haar, YDgCoCg-WSST
        r  = r  + (-b);
        b  = b  + round(r / 2);
        
        g2 = g2 + (-g1);
        g1 = g1 + round(g2 / 2);
        
        g1 = g1 + (-b);
        b  = b  + round(g1 / 2);
    case 53 % 5/3, YDgCoCg-WSST
        p0 = -1/2;
        u0 = 1/4;
        
        r  = r  + round((b + delay1(b) + delay2(b) + delay1(delay2(b))) * p0 / 2);
        b  = b  + round((r + delay3(r) + delay4(r) + delay3(delay4(r))) * u0 / 2);
        
        g2 = g2 + round((g1 + delay3(g1) + delay2(g1) + delay3(delay2(g1))) * p0 / 2);
        g1 = g1 + round((g2 + delay1(g2) + delay4(g2) + delay1(delay4(g2))) * u0 / 2);
        
        g1 = g1 + round((b + delay1(b)) * p0);
        b  = b  + round((g1 + delay3(g1)) * u0);
    case 97 % 9/7, YDgCoCg-WSST
        p0 = -1.58613434205992;
        u0 = -0.05298011857295;
        p1 = 0.882911075530940;
        u1 = 0.443506852043967;
        
        r  = r + round((b + delay1(b) + delay2(b) + delay1(delay2(b))) * p0 / 2);
        b  = b + round((r + delay3(r) + delay4(r) + delay3(delay4(r))) * u0 / 2);
        r  = r + round((b + delay1(b) + delay2(b) + delay1(delay2(b))) * p1 / 2);
        b  = b + round((r + delay3(r) + delay4(r) + delay3(delay4(r))) * u1 / 2);
        
        g2 = g2 + round((g1 + delay3(g1) + delay2(g1) + delay3(delay2(g1))) * p0 / 2);
        g1 = g1 + round((g2 + delay1(g2) + delay4(g2) + delay1(delay4(g2))) * u0 / 2);
        g2 = g2 + round((g1 + delay3(g1) + delay2(g1) + delay3(delay2(g1))) * p1 / 2);
        g1 = g1 + round((g2 + delay1(g2) + delay4(g2) + delay1(delay4(g2))) * u1 / 2);
        
        g1 = g1 + round((b + delay1(b)) * p0);
        b  = b  + round((g1 + delay3(g1)) * u0);
        g1 = g1 + round((b + delay1(b)) * p1);
        b  = b  + round((g1 + delay3(g1)) * u1);
end
y  = b;
dg = g2;
co = r;
cg = g1;

function x = delay1(x)
% up z0
x = [x(1, :); x(1:end-1, :)];

function x = delay2(x)
% left z1
x = [x(:, 1), x(:, 1:end-1)];

function x = delay3(x)
% bottom z0^-1
x = [x(2:end, :); x(end, :)];

function x = delay4(x)
% right z1^-1
x = [x(:, 2:end), x(:, end)];